import asyncio
import functools
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import AsyncIterator, NoReturn, Optional

from aioudp import connection


@dataclass
class ClientProtocol(asyncio.DatagramProtocol):
    msg_queue: "asyncio.Queue[Optional[bytes]]"
    transport: Optional[asyncio.BaseTransport] = None

    # def connection_made(self, transport: "asyncio.BaseTransport") -> None:
    #     self.transport = transport

    def datagram_received(self, data: bytes, _: connection.AddrType) -> None:
        self.msg_queue.put_nowait(data)

    def error_received(self, exc: Exception) -> NoReturn:
        # Haven't figured out why this can happen
        raise exc

    def connection_lost(self, exc: Optional[Exception]) -> None:
        # Haven't figured out why this can happen
        if exc is not None:
            raise exc
        self.msg_queue.put_nowait(None)


@asynccontextmanager
async def connect(host: str, port: int) -> AsyncIterator[connection.Connection]:
    """Connects to a UDP server.

    .. seealso::

        :func:`serve`

    Args:
        host (str): The server's host name/address
        port (int): The server's port number

    """
    loop = asyncio.get_running_loop()
    msgs: "asyncio.Queue[Optional[bytes]]" = asyncio.Queue()
    transport: "asyncio.BaseTransport"
    _: "asyncio.BaseProtocol"
    transport, _ = await loop.create_datagram_endpoint(
        lambda: ClientProtocol(msgs),
        remote_addr=(host, port),
        # FIXME: Should use `local_addr` or `remote_addr` be configurable?
    )
    try:
        yield connection.Connection(  # TODO: REFACTOR: minimal args
            send_func=transport.sendto,  # type: ignore
            recv_func=msgs.get,
            is_closing=transport.is_closing,
            get_local_addr=functools.partial(transport.get_extra_info, "sockname"),
            get_remote_addr=functools.partial(transport.get_extra_info, "peername"),
        )
    finally:
        transport.close()



from dataclasses import dataclass
from typing import Awaitable, Callable, Optional, Tuple

from aioudp import exceptions

__all__ = ["Connection"]

AddrType = Tuple[str, int]


@dataclass
class Connection:  # TODO: REFACTOR: minimal args
    """Represents a server-client connection. Do not instantiate manually."""

    send_func: Callable[[bytes], None]
    recv_func: Callable[[], Awaitable[Optional[bytes]]]
    is_closing: Callable[[], bool]
    get_local_addr: Callable[[], AddrType]
    get_remote_addr: Callable[[], Optional[AddrType]]
    closed: bool = False

    @property
    def local_address(self) -> AddrType:
        """Returns the local address of the connection. This is your IP

        .. seealso::
            :meth:`remote_address`

        Returns:
            AddrType: This is a `tuple` containing the hostname and port

        """
        return self.get_local_addr()  # type: ignore  # This is a bug https://github.com/python/mypy/issues/6910

    @property
    def remote_address(self) -> Optional[AddrType]:
        """Returns the remote address of the connection. This is their IP

        .. seealso::
            :meth:`local_address`

        Returns:
            AddrType: This is a `tuple` containing the hostname and port

        """
        return self.get_remote_addr()  # type: ignore  # See above

    async def recv(self) -> bytes:
        """Receives a message from the connection

        Returns:
            bytes: The received `bytes`.

        Raises:
            exceptions.ConnectionClosedError: The connection is closed

        """
        the_next_one = await self.recv_func()  # type: ignore  # See above
        if the_next_one is None:
            assert self.is_closing()  # type: ignore  # See above
            raise exceptions.ConnectionClosedError("The connection is closed")
        return the_next_one

    async def send(self, data: bytes) -> None:
        """Sends a message to the connection.

        .. warning::
            Since this is UDP, there is no guarantee that the message will be sent

        Args:
            data (bytes): The message in bytes to send

        Raises:
            exceptions.ConnectionClosedError: The connection is closed
            ValueError: There is no data to send
        """
        if self.is_closing():  # type: ignore  # See above
            raise exceptions.ConnectionClosedError("The connection is closed")
        if not data:
            raise ValueError("You must send some data")
        self.send_func(data)  # type: ignore  # See above

    def __aiter__(self) -> "Connection":
        return self

    async def __anext__(self) -> bytes:
        try:
            return await self.recv()
        except exceptions.ConnectionClosedError as error:
            raise StopAsyncIteration from error



"""A better API for asynchronous UDP"""

from . import exceptions
from .client import connect
from .connection import Connection
from .server import serve

__all__ = ["connect", "serve", "Connection", "exceptions"]


__version__ = "0.1.0"



"""Exceptions this library might raise"""


class AioUDPError(Exception):
    """Base exception for this library

    All exceptions this library will raise will subclass this
    """


class ConnectionClosedError(AioUDPError):
    """When a connection is closed and you tried to send/recieve a message"""



import asyncio
import functools
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import AsyncIterator, Awaitable, Callable, Dict, NoReturn, Optional

from aioudp import connection


@dataclass
class ServerProtocol(asyncio.DatagramProtocol):
    handler: Callable[[connection.Connection], Awaitable[None]]
    msg_handler: Dict[connection.AddrType, "asyncio.Queue[Optional[bytes]]"] = field(
        default_factory=dict
    )
    transport: Optional[asyncio.DatagramTransport] = None

    def connection_made(self, transport: "asyncio.DatagramTransport") -> None:  # type: ignore
        self.transport = transport

    def datagram_received(self, data: bytes, addr: connection.AddrType) -> None:
        if addr not in self.msg_handler:
            self.msg_handler[addr] = asyncio.Queue()
            assert self.transport is not None
            asyncio.create_task(
                self.handler(  # type: ignore  # See connnection.py
                    connection.Connection(  # TODO: REFACTOR: minimal args
                        send_func=functools.partial(self.transport.sendto, addr=addr),
                        recv_func=self.msg_handler[addr].get,
                        is_closing=self.transport.is_closing,
                        get_local_addr=functools.partial(
                            self.transport.get_extra_info, "sockname"
                        ),
                        get_remote_addr=lambda: addr,
                    )
                )
            ).add_done_callback(lambda _: self.msg_handler.pop(addr, None))
        self.msg_handler[addr].put_nowait(data)

    def error_received(self, exc: Exception) -> NoReturn:
        # Haven't figured out why this can happen
        raise exc

    def connection_lost(self, exc: Optional[Exception]) -> None:
        # Haven't figured out why this can happen
        if exc is not None:
            raise exc
        for key in self.msg_handler:
            self.msg_handler[key].put_nowait(None)


@asynccontextmanager
async def serve(
    host: str, port: int, handler: Callable[[connection.Connection], Awaitable[None]]
) -> AsyncIterator[None]:
    """Runs a UDP server.

    See the docs for an example UDP echo server

    .. seealso::

        :func:`connect`

        :doc:`An example UDP echo server </index>`

    Args:
        host (str): The host name/address to run the server on
        port (int): The port number to run the server on
        handler (Callable[[connection.Connection], Awaitable[None]]): An asynchronous function to handle a request
            It should accept an instance of :class:`connection.Connection`
            and doesn't need to return anything.

    """
    loop = asyncio.get_running_loop()
    transport: "asyncio.BaseTransport"
    _: "asyncio.BaseProtocol"
    transport, _ = await loop.create_datagram_endpoint(
        lambda: ServerProtocol(handler),
        local_addr=(host, port),
        # FIXME: Should use `local_addr` or `remote_addr` be configurable?
    )
    try:
        yield
    finally:
        transport.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aioudp']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'aioudp',
    'version': '0.1.0',
    'description': 'A better API for asynchronous UDP',
    'long_description': '# AioUDP\n\n[![Documentation Status](https://readthedocs.org/projects/aioudp/badge/?version=latest)](https://aioudp.readthedocs.io/en/latest/?badge=latest)\n\n> A better API for asynchronous UDP\n\nA [websockets](https://websockets.readthedocs.io/en/stable/index.html)-like API for [UDP](https://en.wikipedia.org/wiki/User_Datagram_Protocol)\n\nHere\'s an example echo server:\n\n```py\nimport aioudp\nimport asyncio\n\nasync def main():\n    async def handler(connection):\n        async for message in connection:\n            await connection.send(message)\n    async with aioudp.serve("localhost", 9999, handler):\n        await asyncio.Future()  # Serve forever\n\nif __name__ == \'__main__\':\n    asyncio.run(main())\n```\n\nAnd a client to connect to the server:\n\n```py\nimport aioudp\nimport asyncio\n\nasync def main():\n    async with aioudp.connect("localhost", 9999) as connection:\n        await connection.send(b"Hello world!")\n        assert await connection.recv() == b"Hello world!"\n\nif __name__ == \'__main__\':\n    asyncio.run(main())\n```\n',
    'author': 'Bryan Hu',
    'author_email': 'bryan.hu.2020@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ThatXliner/aioudp',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
